namespace eval VELaSSCo {
    variable _dir
    variable _images
    variable _tunnel_used 0
    variable _sessionID   ""
    variable _modelID     ""
    variable _modelInfo
    variable _status_wid  ""
    variable _login_icon_wid     ""
    variable _icon_do_connection connect_creating.png
    variable _icon_connected     connect_established.png
    variable _ogl
    # maximum vervosity
    variable _verbose_level   9
    # verbose levels:
    # 0 - Nothing
    # 1 - show queries issued with short response

    # utils, Ini & End
    proc GetFullImageName { filename} {
	variable _dir
	return [ file join $_dir images $filename]
    }
    proc GetImage { filename} {
	variable _dir
	variable _images
	if { ![ info exist _images($filename)]} {
	    set _images($filename) [ image create photo -file \
					 [ file join $_dir images $filename]]
	} 
	return $_images($filename)
    }

    proc Ini { dir} {
	variable _dir

	set _dir $dir

	set VersionRequired "12.0.0"
	set comp -1
	catch { 
	    set comp [ GiDVersionCmp $VersionRequired]
	}

	# doing package require ...
	if { [ string first $_dir $::auto_path] == -1} {
	    lappend ::auto_path ${_dir}
	}
	set err_txt ""
	set err [ catch {
	    set ret [ package require gid_velassco]
	    # W "package require gid_velassco = $ret"
	} err_txt ]
	if { $err} {
	    # problems loading the package ( i.e. library)
	    W $err_txt
	    if { $::tcl_platform(platform) == "windows"} {
		set suggestion "\nThe package has been compiled with\n"
		append suggestion "Microsoft Visual C++ 2012\n"
		append suggestion "May be the solution is to install the\n"
		append suggestion "Microsoft Cisual C++ 2012 Redistributable (x64)\n"
		append suggestion "which can be found at [ file join $_dir bin vcredist_x64.exe]"
		W $suggestion
		WarnWin $suggestion
	    } else {
		W "Dependencies:"
		set lib_name [ file join $_dir gid_velassco [ Get_Library_gid_velassco_name gid_velassco]]
		W [ exec ldd $lib_name]
		set suggestion "\nMay be the solution is to install\n"
		append suggestion "boost >= 1.58\n"
		append suggestion "thrift >= 0.92\n"
		append suggestion "libevent >= 2.0"
		W $suggestion
	    }
	}
	
	return
    }

    proc End { args } {
	DoLogout
    }

    proc WaitState { } {
	::GidUtils::WaitState .gid
    }
    
    proc EndWaitState { } {
	::GidUtils::EndWaitState .gid
    }

    proc Print { msg } {
	# W $msg
	GidUtils::SetWarnLine $msg
    }

    proc Verbose { level msg} {
	variable _verbose_level
	if { $level < $_verbose_level} {
	    Print $msg
	}
    }

    proc Debug { msg } {
	# W $msg
	GidUtils::SetWarnLine $msg
    }

    # functionalities / Vqueries
    proc About { } {
	set package_version "N/A"
	catch {
	    set package_version [ package require gid_velassco]
	}
	# InfoWin "About VELaSSCo:\n\nUsing gid_velassco v. $package_version\nwith support for\n[ gid_velassco help]\n\nPlease visit http://www.velassco.eu"

	set w .gid.wVelasscoAbout
	if { [ winfo exists $w]} {
	    destroy $w
	}
	toplevel $w
	wm title $w [_ "About VELaSSCo..."]
	CreateDecorationWindow $w  $w.top $w.bot\
	    [ list [_ "Ok"]] \
	    [ list [ list destroy $w]] \
	    [ list <Return>]
	ttk::frame $w.top.header
	ttk::label $w.top.header.l1 -image [ VELaSSCo::GetImage VelasscoBig67x48.png]
	ttk::label $w.top.header.l2 -text [_ "About VELaSSCo"]
	grid $w.top.header.l1 $w.top.header.l2 -padx 4 -pady 4 -sticky ew

	ttk::frame $w.top.about
	ttk::label $w.top.about.l0 -text "Using gid_velassco v. $package_version"
	ttk::label $w.top.about.l1 -text "with support for:"
        ttk::scrollbar $w.top.about.scrolly -command "$w.top.about.t yview" -orient vertical
	text $w.top.about.t -wrap word -width 20 \
	    -yscrollcommand "$w.top.about.scrolly set" \
	    -border 0 -background [ $w cget -background]
	# -height 8
	VisitWebButton $w.top.about.b http://www.velassco.eu "Please visit www.velassco.eu"

	grid $w.top.about.l0 -sticky w -padx 4 -pady 0 -columnspan 2
	grid $w.top.about.l1 -sticky w -padx 4 -pady 0 -columnspan 2
	grid $w.top.about.t -sticky news -padx 4 -pady 0
	grid $w.top.about.scrolly -row 2 -column 1 -sticky ns
	grid $w.top.about.b -sticky ew -padx 4 -pady 0 -columnspan 2
	grid columnconfigure $w.top.about 0 -weight 1
	grid rowconfigure $w.top.about 2 -weight 1
	# grid rowconfigure $w.top.about 1 -weight 1

	grid $w.top.header -sticky ew
	grid $w.top.about -sticky ew -padx 4 -pady 4
	grid columnconfigure $w.top 0 -weight 1
	grid rowconfigure $w.top 1 -weight 1

	# get rid of gid_velassco ... ?parameters?
	regsub -all { \| } [ lrange [ gid_velassco help] 1 end-1] \n txt
	$w.top.about.t insert end $txt
	# $w.top.about.t insert end ""
	# $w.top.about.t see end
	$w.top.about.t configure -state disabled
    }

    proc OpenModel { { w .gid.wVelasscoOpenModel}} {
	variable _sessionID
	variable _modelID
	variable _modelInfo

	if { ![ CheckLogin]} { return}
	# Debug "VQuery OpenModel( $::VELaSSCo(SelectedModel))"
	if { "$::VELaSSCo(SelectedModel)" != ""} {
	    # should verify if there is a model already opened
	    if { "$_modelID" != ""} {
		CloseModel
	    }
	    set err_txt "Ok"
	    set err [ catch {
		set _modelID [ gid_velassco OpenModel $_sessionID $::VELaSSCo(SelectedModel) rw]
	    } err_txt]
	    Verbose 1 "gid_velassco OpenModel $_sessionID $::VELaSSCo(SelectedModel) rw = $_modelID (error = err - $err_txt)"
	    if { !$err} {
		wm title .gid $::VELaSSCo(SelectedModel)
		set _modelInfo(unique_name) $::VELaSSCo(SelectedModel)
	    }
	}
    }

    proc CloseModel {} {
	variable _sessionID
	variable _modelID
	variable _modelInfo
	if { ![ CheckLogin]} { return}
	if { "$_modelID" != ""} {
	    set err_txt "Ok"
	    set err [ catch {
		gid_velassco CloseModel $_sessionID $_modelID
	    } err_txt]
	    Verbose 1 "gid_velassco CloseModel $_sessionID $_modelID = $err - $err_txt"
	    set _modelID ""
	    OGLunregisterAndDeleteLists
	    array unset _modelInfo
	}
    }

    proc DoSelectModel { w} {
	variable _sessionID
	variable _modelID
	
	if { $::VELaSSCo(SelectedModel) != ""} {
	    destroy $w
	    OpenModel
	    Verbose 1 "selected model = modelID = $_modelID"
	} else {
	    ErrorWin [_ "No model has been selected"]
	}
    }

    proc DoUpdateThumbImage { wThumb img} {
	if { [ winfo exists $wThumb]} {
	    $wThumb configure -text $img
	}
    }

    proc GetAndUpdateThumbnail { x y w tbl wThumb} {

	# set tablelist::x [expr {$x + [winfo x $w]}]
	# set tablelist::y [expr {$y + [winfo y $w]}]
	# set cell [$tbl nearestcell $tablelist::x $tablelist::y]
	# foreach {row col} [ split $cell ,] break    

	set setsel [$tbl curselection]

	set model_unique_name "VELaSSCo_Models:*:*"
	if { $setsel == {} } {
	    tk_dialogRAM $w.terr error [_ "Sorry, no model selected."] error 0 [_ "OK"]
	} else {
	    if { [ llength $setsel] == 1} {
		set row_values [ $tbl get $setsel]
		set model_name [ lindex $row_values 0]
		set model_path [ lindex $row_values 1]
		set model_loc [ lindex $row_values 2]
		# location can be of the form HBase:TableName or EDM
		if { [ string compare -nocase "EDM" $model_loc] == 0 } {
		    # it matches "EDM"
		    set model_unique_name $model_name
		} else {
		    set colon_idx [ string first ":" $model_loc]
		    if { $colon_idx != -1 } {
			set model_unique_name "[ string range $model_loc [ expr $colon_idx + 1] end]:$model_path:$model_name"
		    } else {
			set model_unique_name "$model_path:$model_name"
		    }
		}
	    }
	}
	# at the moment no GetThumbnailOfModel query
	# Debug "VQuery GetThumbnailOfModel( $model_unique_name)"
	# after 500 [ list VELaSSCo::DoUpdateThumbImage $wThumb [_ "Showing thumbnail for model '%s'" $model_unique_name]]
	set ::VELaSSCo(SelectedModel) $model_unique_name
    }

    proc FillModelTable { tbl} {
	variable _sessionID

	$tbl delete 0 end

	set group $::VELaSSCo(group_qualifier)
	set name $::VELaSSCo(name_pattern)
	set ret [ gid_velassco GetListOfModels $_sessionID $group $name]
	# Debug "gid_velassco GetListOfModels $_sessionID $group $name = \n$ret"

	# filling list
	set lst_names {}
	set lst_fp {}
	set lst_model_ids {}
	set lst_locations {}
	
	# we skip the first line: "NumberOfModels: N"
	set lst_models [ lrange [ split $ret "\n"] 1 end]
	foreach entry $lst_models {
	    set keyword [ lindex $entry 0]
	    set value [ lrange $entry 1 end]
	    switch $keyword {
		"Name:" {
		    lappend lst_names $value
		}
		"FullPath:" {
		    lappend lst_fp $value
		}
		"ModelID:" {
		    lappend lst_model_ids $value
		}
		"Location:" {
		    lappend lst_locations $value
		}
	    }
	}
	foreach name $lst_names fp $lst_fp lc $lst_locations modelID $lst_model_ids {
	    $tbl insert end [ list $name $fp $lc $modelID]
	    # after 100
	}
    }

    proc SelectModel { { w .gid.wVelasscoSelectModel}} {
	if { ![ CheckLogin]} { return}
	if { [ winfo exists $w]} {
	    destroy $w
	}
	toplevel $w
	wm title $w [_ "VELaSSCo model selection..."]
	# CreateApplyCloseDecorationWindow $w $w.top $w.bot \
	    # 	{WarnWin Accept} \
	    # 	"destroy $w"
	CreateDecorationWindow $w  $w.top $w.bot\
	    [ list [_ "Open"] [_ "Cancel"]] \
	    [ list [ list VELaSSCo::DoSelectModel $w] [ list destroy $w]] \
	    [ list <Return> <Escape>]
	ttk::frame $w.top.header
	ttk::label $w.top.header.l1 -image [ VELaSSCo::GetImage VelasscoBig67x48.png]
	ttk::label $w.top.header.l2 -text [_ "VELaSSCo model selection"]
	grid $w.top.header.l1 $w.top.header.l2 -padx 4 -pady 4 -sticky ew

	package require tablelist_tile
	ttk::frame $w.top.selection

	ttk::frame $w.top.selection.filter
	ttk::label $w.top.selection.filter.lgroup -text [_ "Group qualifier:"]
	ttk::entry $w.top.selection.filter.egroup -width 20 -textvariable ::VELaSSCo(group_qualifier)
	ttk::label $w.top.selection.filter.lpattern -text [_ "Name pattern:"]
	ttk::entry $w.top.selection.filter.epattern -width 20 -textvariable ::VELaSSCo(name_pattern)
	ttk::button $w.top.selection.filter.brefresh -text [_ "Refresh table"]
	grid $w.top.selection.filter.lgroup $w.top.selection.filter.egroup \
	     $w.top.selection.filter.lpattern $w.top.selection.filter.epattern \
	    $w.top.selection.filter.brefresh \
	    -padx 4 -pady 4 -sticky e
	grid configure  $w.top.selection.filter.egroup -sticky w
	grid configure  $w.top.selection.filter.epattern -sticky w
	grid configure  $w.top.selection.filter.brefresh -sticky w

	set tbl_sw [ ScrolledWindow $w.top.selection.sw -auto both -relief solid -borderwidth 0]
	set tbl $w.top.selection.sw.tbl
	tablelist::tablelist $tbl \
	    -stretch all -background white \
	    -columns [ list 20 [_ "Name"] 0 [_ "Full path"] 0 [_ "Location"] 32 [_ "Model ID"]]
	    # -columns [ list 30 [_ "Name"] 0 [_ "Full path"] 20 [_ "Validation status"] 20 [_ "Access"]]

	$w.top.selection.sw setwidget $tbl

	ttk::label $w.top.selection.thumb -text [_ "Thumbnail image of selected model"]
	bind $tbl <<TablelistSelect>> [ list VELaSSCo::GetAndUpdateThumbnail %x %y %W $tbl $w.top.selection.thumb]

	bind [$tbl bodytag] <Double-Button-1> [ list VELaSSCo::DoSelectModel $w]

	grid $w.top.selection.filter -sticky ew
	# grid $w.top.selection.tbl -padx 4 -pady 4 -sticky news
	grid $w.top.selection.sw -padx 4 -pady 4 -sticky news
	grid $w.top.selection.thumb  -padx 4 -pady 4 -sticky ew

	grid columnconfigure $w.top.selection 0 -weight 1
	grid rowconfigure $w.top.selection 1 -weight 1

	grid $w.top.header -sticky ew
	grid $w.top.selection -sticky ewns
	grid columnconfigure $w.top 0 -weight 1
	grid rowconfigure $w.top 1 -weight 1
	wm geometry $w 990x325

	set ::VELaSSCo(group_qualifier) "*"
	set ::VELaSSCo(name_pattern) "*"

	$w.top.selection.filter.brefresh configure -command [ list ::VELaSSCo::FillModelTable $tbl] 
	FillModelTable $tbl
    }

    proc ShowModelInfo {} {
	variable _sessionID
	variable _modelID
	if { ![ CheckLoginAndModelID]} { return}
	set doing ""
	set err [ catch {
	    # testing GetListOfAnalyses
	    set doing "gid_velassco GetListOfAnalyses $_sessionID $_modelID"
	    set lst_analyses [ gid_velassco GetListOfAnalyses $_sessionID $_modelID]
	    set analysis [ lindex $lst_analyses 0]
	    W "gid_velassco GetListOfAnalyses $_sessionID $_modelID = "
	    W $lst_analyses
	    W "  selected = $analysis"
	    W \n
	    # testing GetListOfSteps
	    set doing "gid_velassco GetListOfSteps $_sessionID $_modelID $analysis"
	    set lst_steps [ gid_velassco GetListOfSteps $_sessionID $_modelID $analysis]
	    set step [ lindex $lst_steps 0]
	    W "gid_velassco GetListOfSteps $_sessionID $_modelID $analysis = "
	    set num_steps [ llength $lst_steps]
	    W "Number of Steps = $num_steps"
	    if { $num_steps < 20} {
		W $lst_steps
	    } else {
		# show only some of them
		W "[ lrange $lst_steps 0 7] . . . [ lrange $lst_steps [ expr $num_steps - 8] end]"
	    }
	    W "  selected = $step"
	    W \n
	    # testing GetListOfResults
	    set doing "gid_velassco GetListOfResults $_sessionID $_modelID $analysis $step"
	    set lst_results [ gid_velassco GetListOfResults $_sessionID $_modelID $analysis $step]
	    W "gid_velassco GetListOfResults $_sessionID $_modelID $analysis $step = "
	    W $lst_results
	    W \n
	    
	    # testing GetListOfMeshes
	    if { "$analysis" != "DEM"} {
		# get information about static meshes
		set analysis ""
		set step 0.0
	    }
	    set doing "gid_velassco GetListOfMeshes $_sessionID $_modelID $analysis $step"
	    set lst_meshes [ gid_velassco GetListOfMeshes $_sessionID $_modelID $analysis $step]
	    W "gid_velassco GetListOfMeshes $_sessionID $_modelID \"$analysis\" $step = "
	    W $lst_meshes
	    W \n
	} err_txt]
	if { $err} {
	    W "ERROR doing $doing"
	    W $err_txt
	}
    }

    proc DoTunnelLinux {} {
	set velassco_host "$::VELaSSCo(Username)@$::VELaSSCo(Host) port $::VELaSSCo(Port) (tunneling)"
	# set _tunnel_used [ exec ssh -N -L $::VELaSSCo(Tunnel,LocalPort):$::VELaSSCo(Host):$::VELaSSCo(Port) $::VELaSSCo(Tunnel,Username)@$::VELaSSCo(Tunnel,Host) &]
	# needs expect
	set err [ catch {
	    set fo [ file tempfile tmp_file velassco]
	    puts $fo "#!/usr/bin/expect"
	    puts $fo "# Using ssh from expect"
	    puts $fo "log_user 0"
	    puts $fo "spawn ssh -N -L $::VELaSSCo(Tunnel,LocalPort):$::VELaSSCo(Host):$::VELaSSCo(Port) $::VELaSSCo(Tunnel,Username)@$::VELaSSCo(Tunnel,Host)"
	    puts $fo "expect \"sword: \""
	    puts $fo "send \"$::VELaSSCo(Tunnel,Password)\\r\""
	    puts $fo "interact"
	    close $fo
	    file attributes $tmp_file -permissions u+x,go-rwx
	    set _tunnel_used [ exec $tmp_file &]
	    # Debug $_tunnel_used
	    after 1000
	    file delete -force $tmp_file
	} err_txt]
	if { $err} {
	    Print "Could not use expect: '$err_txt', using interactive xterm isntead."
	    set _tunnel_used [ exec xterm -e ssh -N -L $::VELaSSCo(Tunnel,LocalPort):$::VELaSSCo(Host):$::VELaSSCo(Port) $::VELaSSCo(Tunnel,Username)@$::VELaSSCo(Tunnel,Host) &]
	    after 5000
	}
    }

    proc DoTunnelWindows {} {
	variable _dir
	variable _tunnel_used
	
	set velassco_host "$::VELaSSCo(Username)@$::VELaSSCo(Host) port $::VELaSSCo(Port) (tunneling)"
	# set _tunnel_used [ exec ssh -N -L $::VELaSSCo(Tunnel,LocalPort):$::VELaSSCo(Host):$::VELaSSCo(Port) $::VELaSSCo(Tunnel,Username)@$::VELaSSCo(Tunnel,Host) &]
	# needs expect
	set bin_putty [ file join $_dir bin putty.exe]
	set _tunnel_used [ exec "$bin_putty" -N -L $::VELaSSCo(Tunnel,LocalPort):$::VELaSSCo(Host):$::VELaSSCo(Port) $::VELaSSCo(Tunnel,Username)@$::VELaSSCo(Tunnel,Host) &]
	# set putty_cmd "\"$bin_putty\" -N -L $::VELaSSCo(Tunnel,LocalPort):$::VELaSSCo(Host):$::VELaSSCo(Port) $::VELaSSCo(Tunnel,Username)@$::VELaSSCo(Tunnel,Host)"
	# proc ::___dummy__cb {} {
	# }
	# set _tunnel_used [ runExe run $putty_cmd -blocking false -callback ::___dummy__cb -timeout 0]
	after 5000
    }

    proc DoLogin { w} {
	variable _sessionID
	variable _tunnel_used

	if { $_sessionID != ""} {
	    WarnWin [_ "Already logged in!"]
	    return
	}

	if { $::VELaSSCo(Tunnel,enabled)} {

	    if { $::tcl_platform(platform) == "windows"} {
		DoTunnelWindows
	    } else {
		DoTunnelLinux
	    }

	    set gv_id [ gid_velassco UserLogin localhost $::VELaSSCo(Tunnel,LocalPort) $::VELaSSCo(Username) $::VELaSSCo(Password)]

	    # when closing the window do a logout ...
	    bind $w <Destroy> "+::VELaSSCo::DoLogout"

	} else {
	    set _tunnel_used 0
	    set gv_id [ gid_velassco UserLogin $::VELaSSCo(Host) $::VELaSSCo(Port) $::VELaSSCo(Username) $::VELaSSCo(Password)]
	}
	set _sessionID $gv_id

	ShowLoginStatus
    }

    proc DoLogout {} {
	variable _sessionID
	variable _tunnel_used
	if { $_sessionID != ""} {
	    set ret ""
	    catch { 
		set ret [ gid_velassco UserLogout $_sessionID]
	    }
	    if { $_tunnel_used != 0} {
		# kill tunnel
		if { $::tcl_platform(platform) == "windows"} {
		    # runExe stop $_tunnel_used
		    package require gid_utils
		    gid_utils::end_process $_tunnel_used
		} else {
		    exec kill $_tunnel_used
		}
	    }
	    Verbose 1 "gid_velassco UserLogout $_sessionID = $ret"
	}
	set _sessionID ""
	set _tunnel_used 0
	ShowLoginStatus
    }

    proc DoStatusDB {} {
	variable _sessionID
	if { ![ CheckLogin]} { return}
	InfoWin [_ "gid_velassco GetStatusDB $_sessionID = \n%s" \
		     [ gid_velassco GetStatusDB $_sessionID]]
    }

    proc EnableDisableTunnelWidgets { lst_w_tunnel} {
	foreach w $lst_w_tunnel {
	    set state disabled
	    if { $::VELaSSCo(Tunnel,enabled)} {
		set state normal

		# set new default parameters
		set ::VELaSSCo(Host) pez001
		set ::VELaSSCo(Port) 26267
	    }
	    $w configure -state $state
	}
    }

    proc ShowLoginStatus {} {
	variable _sessionID
	variable _status_wid
	variable _login_icon_wid
	variable _icon_do_connection
	variable _icon_connected
	
	if { ( $_status_wid != "") && [ winfo exists $_status_wid]} {
	    set color green
	    set txt "Logged in"
	    if { $_sessionID == ""} {
		set color red
		set txt "Not logged in"
	    }
	    $_status_wid configure -background $color
	    $_status_wid.txt configure -text $txt -background $color -foreground white
	}

	if { ( $_login_icon_wid != "") && [ winfo exists $_login_icon_wid]} {
	    set img $_icon_connected
	    if { $_sessionID == ""} {
		set img $_icon_do_connection
	    }
	    $_login_icon_wid configure -image [ VELaSSCo::GetImage $img]
	}
    }

    proc CheckLogin {} {
	variable _sessionID

	set ret 1
	if { $_sessionID == ""} {
	    WarnWin [_ "Please, log in first."]
	    set ret 0
	} else {
	    # may be some test like GetStatusDB to see if connection is alive ...
	}
	return $ret
    }
    proc CheckLoginAndModelID {} {
	variable _modelID

	if { ![ CheckLogin]} { return 0}
	set ret 1
	if { $_modelID == ""} {
	    WarnWin [_ "Please, open a model first."]
	    set ret 0
	} else {
	    # may be some test like GetStatusDB to see if connection is alive ...
	}
	return $ret
    }

    proc OGLdrawLists {} {
	variable _ogl
	if { [ info exists _ogl]} {
	    foreach name [ array names _ogl list,*] {
		GiD_OpenGL draw -calllist $_ogl($name)
	    }
	}
    }
    proc OGLregisterDraw {} {
	variable _ogl
	if { [ info exists _ogl]} {
	    set _ogl(register_id) [ GiD_OpenGL register ::VELaSSCo::OGLdrawLists]
	}
    }
    proc OGLunregisterAndDeleteLists {} {
	variable _ogl
	if { [ info exists _ogl]} {
	    if { [ info exists _ogl(register_id)] && ( $_ogl(register_id) != -1)} {
		GiD_OpenGL unregister $_ogl(register_id)
		set _ogl(register_id) -1
	    }
	    # now deleting lists
	    foreach name [ array names _ogl list,] {
		GiD_OpenGL draw -deletelists $_ogl($name) 1
	    }
	}
    }

    proc OGLcreateBBoxList { bbox} {
	variable _ogl
	
	# create list
	set _ogl(list,bbox) [ GiD_OpenGL draw -genlists 1]
	foreach var [ list min_x min_y min_z max_x max_y max_z] val $bbox {
	    set $var $val
	}
	
	set vertices(0) [ list $min_x $min_y $min_z]
	set vertices(1) [ list $max_x $min_y $min_z]
	set vertices(2) [ list $max_x $max_y $min_z]
	set vertices(3) [ list $min_x $max_y $min_z]
	set vertices(4) [ list $min_x $min_y $max_z]
	set vertices(5) [ list $max_x $min_y $max_z]
	set vertices(6) [ list $max_x $max_y $max_z]
	set vertices(7) [ list $min_x $max_y $max_z]
	set lst_lines [ list \
			    [ list 0 1] [ list 1 2] [ list 2 3] [ list 3 0] \
			    [ list 0 4] [ list 1 5] [ list 2 6] [ list 3 7] \
			    [ list 4 5] [ list 5 6] [ list 6 7] [ list 7 4]]

	GiD_OpenGL draw -newlist $_ogl(list,bbox) compile
	# somewhat dark magenta
	GiD_OpenGL draw -color "0.69841 0.190464 0.69841"
	GiD_OpenGL draw -begin lines
	foreach ln $lst_lines {
	    foreach nd $ln {
		GiD_OpenGL draw -vertex $vertices($nd)
	    }
	}
	# end lines
	GiD_OpenGL draw -end
	# end of  list
	GiD_OpenGL draw -endlist

	#register OpenGL proc if not already registered
	OGLregisterDraw
    }

    # useof: can be "GEOMETRYUSE", "MESHUSE", "POSTUSE" or "GRAPHUSE". 
    proc TclCalcModelBoundaries { useof} {
	variable _modelInfo
	if { [ info exists _modelInfo] && [ info exists _modelInfo(bbox)]} {
	    return $_modelInfo(bbox)
	} else {
	    return [ list 1e+38 1e+38 1e+38 -1e+38 -1e+38 -1e+38]
	}
    }

    proc DoGetBoundingBox {} {
	variable _sessionID
	variable _modelID
	variable _modelInfo

	if { ![ CheckLoginAndModelID]} { return}
	WaitState
	Verbose 1 "calling gid_velassco GetBoundingBox $_sessionID $_modelID"
	set modelBBox [ gid_velassco GetBoundingBox $_sessionID $_modelID]
	# set modelBBox [ list 0.0 0.0 0.0 1.0 1.0 1.0]
	Verbose 1 "selected model = BBox = $modelBBox"
	
	# register a drawopengl procedure to draw the bounding box
	if { [ llength $modelBBox] == 6} {
	    OGLcreateBBoxList $modelBBox
	}
	set _modelInfo(bbox) $modelBBox
	GiD_Process 'Zoom Frame
	EndWaitState
    }

    proc Login { { w_bar_icon ""} { w .gid.wVelasscoLogin}} {
	variable _status_wid
	variable _login_icon_wid
	if { $w_bar_icon != ""} {
	    set _login_icon_wid $w_bar_icon
	}

	if { [ winfo exists $w]} {
	    destroy $w
	}
	toplevel $w
	wm title $w [_ "VELaSSCo access window..."]
	# CreateApplyCloseDecorationWindow $w $w.top $w.bot \
	    # 	{WarnWin Accept} \
	    # 	"destroy $w"
	CreateDecorationWindow $w  $w.top $w.bot\
	    [ list [_ "Login"] [_ "Status"] [_ "Logout"]] \
	    [ list [ list VELaSSCo::DoLogin $w] \
		  VELaSSCo::DoStatusDB VELaSSCo::DoLogout] \
	    [ list <Return> ""]
	ttk::frame $w.top.header
	ttk::label $w.top.header.l1 -image [ VELaSSCo::GetImage VelasscoBig67x48.png]
	ttk::label $w.top.header.l2 -text [_ "VELaSSCo login"]
	grid $w.top.header.l1 $w.top.header.l2 -padx 4 -pady 4 -sticky ew

	ttk::frame $w.top.login
	foreach id [ list 1 2 3 4] \
	    var_name [ list "Username" "Password" "Host" "Port"] \
	    txt [ list [_ "Username"] [_ "Password"] [_ "VELaSSCo host"] [_ "Port"]] \
	    width [ list 16 16 20 8] {
		ttk::label $w.top.login.l$id -text ${txt}:
		ttk::entry $w.top.login.e$id -width $width -textvariable ::VELaSSCo($var_name)
		grid $w.top.login.l$id $w.top.login.e$id -padx 4 -pady 4 -sticky e
		grid configure  $w.top.login.e$id -sticky w
		if { $var_name == "Password"} {
		    $w.top.login.e$id configure -show *
		}
	    }
	ttk::checkbutton $w.top.login.cbTunel -text [_ "Tunnel through"]: \
	    -variable ::VELaSSCo(Tunnel,enabled)
	grid $w.top.login.cbTunel -padx 4 -pady 4 -sticky w -columnspan 2
	set lst_w_tunnel {}
	foreach id [ list 5 6 7 8] \
	    var_name [ list "Host" "LocalPort" "Username" "Password" ] \
	    txt [ list [_ "Tunnel host"] [_ "local port"] [_ "Username"] [_ "Password"] ] \
	    width [ list 20 8 16 16] {
		ttk::label $w.top.login.l$id -text ${txt}:
		ttk::entry $w.top.login.e$id -width $width -textvariable ::VELaSSCo(Tunnel,$var_name)
		grid $w.top.login.l$id $w.top.login.e$id -padx 4 -pady 4 -sticky e
		grid configure  $w.top.login.e$id -sticky w
		lappend lst_w_tunnel $w.top.login.l$id $w.top.login.e$id
		if { $var_name == "Password"} {
		    $w.top.login.e$id configure -show *
		}
	    }
	$w.top.login.cbTunel configure -command [ list VELaSSCo::EnableDisableTunnelWidgets $lst_w_tunnel]

	frame $w.top.status -height 32 -bd 0
	ttk::label $w.top.status.txt -justify center
	grid $w.top.status.txt -padx 4 -pady 4
	
	set _status_wid $w.top.status

	grid columnconfigure $w.top.login 0 -weight 1
	grid columnconfigure $w.top.login 1 -weight 1
	grid rowconfigure $w.top.login 1 -weight 1
	grid rowconfigure $w.top.login 2 -weight 1

	grid columnconfigure $w.top.status 0 -weight 1
	grid rowconfigure $w.top.status 0 -weight 1

	grid $w.top.header -sticky ew
	grid $w.top.login -sticky ew
	grid $w.top.status -sticky ew
	grid columnconfigure $w.top 0 -weight 1
	grid rowconfigure $w.top 1 -weight 1

	# default values
	# foreach var_name [ list "Username" "Password" "Host" "Port"] \
	#     value [ list miguel 1234velassco pez001 9990] {
	# 	set ::VELaSSCo($var_name) $value
	#     }
	# set ::VELaSSCo(Tunnel,enabled) 1
	# foreach var_name [ list "Host" "LocalPort" "Username" "Password" ] \
	#     value [ list acuario.cimne.upc.edu 9990 miguel 1234] {
	# 	set ::VELaSSCo(Tunnel,$var_name) $value
	#     }
	# set ::VELaSSCo(SelectedModel) ""
	foreach var_name [ list "Username" "Password" "Host" "Port"] \
	    value [ list andreas 1234 localhost 9990] {
		set ::VELaSSCo($var_name) $value
	    }
	set ::VELaSSCo(Tunnel,enabled) 0
	foreach var_name [ list "Host" "LocalPort" "Username" "Password" ] \
	    value [ list acuario.cimne.upc.edu 9990 miguel ****] {
		set ::VELaSSCo(Tunnel,$var_name) $value
	    }
	set ::VELaSSCo(SelectedModel) ""

	# enable/disable tunel widgets ...
	VELaSSCo::EnableDisableTunnelWidgets $lst_w_tunnel
	ShowLoginStatus
    }

    proc modifyGiDmenus { } {
	set idx [ GiDMenu::GetOptionIndex Files [ list "Recent post files"] POST]
	if { $idx == -1 } {
	    set idx end
	}
	# set icon [ ::VELaSSCo::GetImage Velassco16.png]
	set icon [ ::VELaSSCo::GetFullImageName Velassco16.png]
	# skip the --- separator just after Recent
	incr idx
	GiDMenu::InsertOption Files [ list "Connect to VELaSSCo ..."] $idx POST ::VELaSSCo::Login \
	    "" $icon "insertafter" _
	incr idx
	GiDMenu::InsertOption Files [ list "Open VELaSSCo model ..."] $idx POST ::VELaSSCo::SelectModel \
	    "" $icon "insertafter" _
	incr idx
	GiDMenu::InsertOption Files [ list ---] $idx "POST" "" "" "" "insertafter"
    }

    proc velasscoIconBarProc { { what DEFAULT} { img_dir ""}} {
	variable _dir
	variable _icon_do_connection
	variable _icon_connected
	global velasscoIconBarBitmapsNames velasscoIconBarBitmapsCommands velasscoIconBarBitmapsHelp
	
	set velasscoIconBarBitmapsNames(0) "Velassco24.png --- $_icon_do_connection --- open2.gif info.png volume.png --- help.gif"
	
	set velasscoIconBarBitmapsCommands(0) [ list {-np- ::VELaSSCo::About} "" \
						    {-np- ::VELaSSCo::Login %W} "" \
						    {-np- ::VELaSSCo::SelectModel} {-np- ::VELaSSCo::ShowModelInfo} \
						    {-np- ::VELaSSCo::DoGetBoundingBox} "" \
						    {-np- ::VELaSSCo::DoStatusDB} ]
	set velasscoIconBarBitmapsHelp(0) [ list "About VELaSSCo" "" \
						"Connect to VELaSSco server" "" \
						"Open a Model" "Show Model information" \
						"GetBoundingBox" "" \
						"VELaSSCo status"]
	
	CreateOtherBitmaps velasscoIconBar "VELaSSCo icon bar" velasscoIconBarBitmapsNames velasscoIconBarBitmapsCommands \
	    velasscoIconBarBitmapsHelp [ file join $_dir images] velasscoIconBarBitmaps $what Post
	
	AddNewToolbar velasscoIcon_bar PostVelasscoIconBarWindowGeom ::VELaSSCo::velasscoIconBarProc "VELaSSCo icon bar"
    }

}

proc InitGIDProject { dir} {
    ::VELaSSCo::Ini $dir

    # modify menus
    ::VELaSSCo::modifyGiDmenus

    # change to postprocess
    GiD_Process Mescape Postprocess

    # Create a new toolbar
    ::VELaSSCo::velasscoIconBarProc INSIDELEFT
    after 1000 ::VELaSSCo::velasscoIconBarProc INSIDELEFT
}

proc EndGIDProject { } {
    ::VELaSSCo::End
}
proc EndGIDPostProcess { } {
    ::VELaSSCo::End
}
proc TclCalcModelBoundaries { useof} {
    # useof: can be "GEOMETRYUSE", "MESHUSE", "POSTUSE" or "GRAPHUSE". 
    ::VELaSSCo::TclCalcModelBoundaries $useof
}
